#include "MainWindow.h"
#include "GeneratedFiles/ui_MainWindow.h"

/*!
 * @class MainWindow
 * @brief implements the functionality of the MainWindow.
 *	
 */

/*!
 * @brief constructor, initializes all widgets and connects all slots and signals
 * @param QWidget *parent: parent widget
 *	
 */
MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent), ui(new Ui::MainWindow)
{
	ui->setupUi(this);
	scene = new GraphicsScene;
	width = 5000;
	height = 1500;
	graph = new StreamGraph( width, height);
	scene->setSceneRect(QRectF(0, 0, width, height));
	view = new GraphicsView();
	view->setScene(scene);
	view->setRenderHints(QPainter::Antialiasing);
	ui->graphLayout->addWidget(view);
	tfpop = new TFpopup(this);
	resizeD = new ResizeDialog(this);
	seed = 0;
	showGrid = false;
	screenCount = 0;

    connect(ui->actionClose, SIGNAL(triggered()), this, SLOT(close()));
    connect(ui->actionUpdate, SIGNAL(triggered()), this, SLOT(refresh()));
	connect(ui->actionLoad_Listening_History, SIGNAL(triggered()), this, SLOT(fileOpen()));
	connect(ui->sortComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(sortOrder(int)));
	connect(ui->layoutComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(layoutSelect(int)));
	connect(ui->scaleXslider, SIGNAL(valueChanged(int)), this, SLOT(scaleX(int)));
	connect(ui->scaleYslider, SIGNAL(valueChanged(int)), this, SLOT(scaleY(int)));
	connect(scene, SIGNAL(mousePos(int, Layer)), this, SLOT(showLabel(int, Layer)));
	connect(ui->colorButton, SIGNAL(clicked()), this, SLOT(showTF()));
	connect(tfpop, SIGNAL(sendFunction(float*)), this, SLOT(updateColorPalette(float*)));
	connect(tfpop, SIGNAL(changeColorCriterion(int)), this, SLOT(setColorCriterion(int)));
	connect(ui->anti_box, SIGNAL(toggled(bool)), this, SLOT(toggleAA(bool)));
	connect(ui->grid_box, SIGNAL(toggled(bool)), this, SLOT(toggleGrid(bool)));
	connect(tfpop, SIGNAL(toggled2D(bool)), this, SLOT(toggle2D(bool)));
	connect(ui->screenshot_button, SIGNAL(clicked()), this, SLOT(takeScreen()));
	connect(ui->actionResize_Canvas, SIGNAL(triggered()), this, SLOT(showResizeDialog()));
	connect(resizeD, SIGNAL(newResolution(int, int)), this, SLOT(resizeScene(int, int)));
	connect(ui->y_slider, SIGNAL(valueChanged(int)), this, SLOT(moveYAxis(int)));

	tfpop->transFunc->prepareTF();
}

/*!
 * @brief rounds a float to the nearest integer
 * @param float d: float to round
 * @return int: rounded float
 *	
 */
int round(float d)
{
	return floor(d + 0.5);
}

/*!
 * @brief this SLOT sets labels to the values of the specified layer
 * @param int pos: mouse position on canvas
 * @param Layer lay: layer on mouse positon
 *	
 */
void MainWindow::showLabel(int pos, Layer lay)
{
	if(pos != -1){
		ui->layer_label->setText(lay.name);

		QColor y = lay.color;
		QString color = "background-color: rgb(" + QString::number(y.red()) + "," + QString::number(y.green()) + "," + QString::number(y.blue()) + ")";
		ui->layer_label->setStyleSheet( color );

		int posit = round((pos-graph->startX)/graph->scale);
		ui->total_label->setText("Sum: " + QString::number(lay.sum*0.1));
		ui->week_label->setText("Week " + QString::number(posit) + ": " + QString::number(lay.size[posit]*0.1));

		//debug stuff
		ui->onset_label->setText("onset: " + QString::number(lay.onset));
		ui->vol_label->setText("vol: " + QString::number(lay.volatility));
		ui->end_label->setText("end: " + QString::number(lay.end));
	}
	else{
		ui->layer_label->setText("");
		ui->week_label->setText("");
		ui->total_label->setText("");

		QColor y = QColor(255,255,255);
		QString color = "background-color: rgb(" + QString::number(y.red()) + "," + QString::number(y.green()) + "," + QString::number(y.blue()) + ")";
		ui->layer_label->setStyleSheet( color );
	}
}

/*!
 * @brief this SLOT sets a new sort order in StreamGraph and draws the new scene
 * @param int i: new sort order
 *	
 */
void MainWindow::sortOrder(int i){
	scene->clear();
	graph->setSort(i);
	drawGrid();
	drawScene();
	return;
}

/*!
* @brief this SLOT sets a new layout in StreamGraph and draws the new scene
* @param int i: new sort order
*	
*/
void MainWindow::layoutSelect(int i){
	scene->clear();
	graph->setLayout(i);
	drawGrid();
	drawScene();
	return;
}

/*!
 * @brief destructor
 *	
 */
MainWindow::~MainWindow()
{
    delete ui;
}

/*!
 * @brief generated by QT 
 *	
 */
void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}


/*!
* \brief
* SLOT that closes the application.
*/
void MainWindow::close()
{
     exit(0);
     return;
}

/*!
* \brief
* SLOT that generates a new random StreamGraph
*/
void MainWindow::refresh()
{
	graph->generateGraph(++seed);
	scene->clear();
	tfpop->transFunc->prepareTF();
	drawGrid();
	drawScene();
    return;
}

/*!
 * @brief SLOT that sets a new x axis scale in StreamGraph
 * @param int i: new x axis scale
 *	
 */
void MainWindow::scaleX(int i)
{
	scene->clear();
	graph->setXscale(i);
	drawGrid();
	drawScene();
	return;
}

/*!
* @brief SLOT that sets a new y axis scale in StreamGraph
* @param int i: new y axis scale
*	
*/
void MainWindow::scaleY(int i)
{
	scene->clear();
	graph->setYscale(i);
	drawGrid();
	drawScene();
	return;
}

/*!
 * @brief SLOT that opens a QFileDialog that is used to load a new last.fm csv file. draws new scene afterwards
 *	
 */
void MainWindow::fileOpen()
{
	scene->clear();
	QString fileName = QFileDialog::getOpenFileName(this, tr("Open Listening History"), "/home/", tr("Last.fm History (*.csv)"));
	graph->loadNewGraph(fileName);
	tfpop->transFunc->prepareTF();
	drawGrid();
	drawScene();
	return;
}

/*!
 * @brief SLOT that opens the color palette dialog window
 *	
 */
void MainWindow::showTF()
{
	tfpop->show();
}

/*!
 * @brief SLOT that receives a new color array from the color palette dialog and updates the StreamGraph with it
 * @param float* colors: color array received from color palette dialog
 *	
 */
void MainWindow::updateColorPalette(float* colors)
{
	this->colors = colors;
	scene->clear();
	graph->updateColorPalette(colors);
	drawGrid();
	drawScene();
	return;
}

/*!
 * @brief SLOT that sets a new coloring criterion and updates StreamGraph
 * @param int i: new coloring criterion
 *	
 */
void MainWindow::setColorCriterion(int i)
{
	graph->criterion = i;
	updateColorPalette(colors);
}

/*!
 * @brief SLOT that sets the QGraphicsView's anti aliasing on or off
 * @param bool on: toggles AA
 *	
 */
void MainWindow::toggleAA(bool on)
{
	if(on) view->setRenderHint(QPainter::Antialiasing, true);
	else view->setRenderHint(QPainter::Antialiasing, false);
}

/*!
 * @brief SLOT that shows/hides the grid in the scene
 * @param bool on: grid on/off
 *	
 */
void MainWindow::toggleGrid(bool on)
{
	showGrid = on;
	scene->clear();
	drawGrid();
	drawScene();
}

/*!
 * @brief draws the grid in the scene if showGrid == true
 *	
 */
void MainWindow::drawGrid()
{
	if(showGrid){
		//if(graph->arrayLength>0){
		//	top.setX(graph->startX - 80);
		//	top.setY(graph->yMax+25);
		//	QGraphicsTextItem* txt = scene->addText("Weeks:", QFont("Garamond", 14));
		//	txt->setPos(top);
		//}
		for(int i=0; i<graph->arrayLength; i++){
			bottom.setX(graph->startX + graph->scale*(i));
			top.setX(graph->startX + graph->scale*(i));
			bottom.setY(graph->yMin-10);
			top.setY(graph->yMax+10);
			scene->addLine(QLine(top, bottom));
			QGraphicsTextItem* txt = scene->addText(QString::number(i), QFont("Garamond", 14));
			if(i<10) txt->setPos(top.x()-9, top.y()+15);
			else if(i>=10) txt->setPos(top.x()-12, top.y()+15);
		}
	}
}

/*!
 * @brief reads each layers path from StreamGraph and adds it to the scene
 *	
 */
void MainWindow::drawScene()
{
	for (int i=0; i<graph->numberOfLayers; i++){
		QGraphicsItem *item = scene->addPath(graph->layers[i].path, QPen(graph->layers[i].color, 1, Qt::SolidLine, Qt::RoundCap), QBrush(graph->layers[i].color, Qt::SolidPattern));
		scene->pathLookupMap[item] = graph->layers[i];
	}
}

/*!
 * @brief switches two dimensional coloring on/off
 * @param bool on: 2D coloring on/off
 *	
 */
void MainWindow::toggle2D(bool on)
{
	scene->clear();
	graph->twoD = on;
	graph->updateColorPalette(colors);
	drawGrid();
	drawScene();
}

void MainWindow::takeScreen()
{
	QImage img(width,height,QImage::Format_ARGB32_Premultiplied);
	QPainter p(&img);
	scene->render(&p);
	p.end();
	screenCount++;
	img.save(QString::number(screenCount)+"screen.png");
}

void MainWindow::showResizeDialog()
{
	resizeD->show();
}

void MainWindow::resizeScene(int w, int h)
{
	resizeD->hide();
	width = w;
	height = h;
	graph->dimX = w;
	graph->dimY = h;
	scene->setSceneRect(QRectF(0, 0, width, height));
}

void MainWindow::moveYAxis(int i)
{
	scene->clear();
	graph->dimY = height + i - 500;
	graph->updateGraph();
	drawGrid();
	drawScene();
}