
#include <QtGui>

#include "StreamGraph.h"

/*!
 * @class StreamGraph
 * @author Johannes Sorger
 * @brief this class calculates the QPainterPaths and QColor for each layer that will be added to the scene.
 *	
 */

/*!
 * @brief streamgraph constructor

	initializes all necessary variables.

 * @param int dimX: width of canvas
 * @param int dimY: height of canvas
 * @param QWidget *parent: parent widget
 *	
 */
StreamGraph::StreamGraph( int dimX, int dimY, QWidget *parent) : QWidget(parent)
{
	this->dimX = dimX;
	this->dimY = dimY;
    scale = 30;
	yScale = 1;
    j=0;
	debugLength = 0;
	sortOrder = 0;
	layoutSelect = 0;
	criterion = 0;
	numberOfLayers = 0;
	arrayLength = 0;
	twoD = false;
}

/*!
 * @brief generates new random layers according to a seed value and amount and length values
 * @param int seed: random seed for generation
 *	
 */
void StreamGraph::generateGraph(int seed){
	data = DataGenerator(seed);
	//TODO: einstellbar machen
	numberOfLayers = 7;
	arrayLength = 20;
	layers = data.make(numberOfLayers, arrayLength);
	updateColorPalette(colors);
	setSort(sortOrder);
}

/*!
 * @brief sorts layers in the specified order
 * @param int order: specifies order
 *	
 */
void StreamGraph::setSort(int order){
	sortOrder = order;
	switch (sortOrder)
	{
	case 0:
		layers = onsetSort.sortLayers(layers, numberOfLayers);
		break;
	case 1:
		layers = basicOnset.sortLayers(layers, numberOfLayers);
		break;
	case 2:
		layers = volSort.sortLayers(layers, numberOfLayers);
		break;
	case 3:
		layers = invVolSorting.sortLayers(layers, numberOfLayers);
		break;
	case 4:
		layers = noSorting.sortLayers(layers, numberOfLayers);
		break;
	}
	
	setLayout(layoutSelect);
}

/*!
 * @brief sets the layout of sorted layers => thus generates baselines necessary for path calculation
 * @param int ls: specifies the selected layout
 *	
 */
void StreamGraph::setLayout(int ls){
	layoutSelect = ls;
	switch (layoutSelect)
	{
	case 0:
		strLayout.layout(layers, numberOfLayers, arrayLength);
		break;
	case 1:
		minWigLayout.layout(layers, numberOfLayers, arrayLength);
		break;
	case 2:
		layout.layout(layers, numberOfLayers, arrayLength);
		break;
	case 3:
		thrvLayout.layout(layers, numberOfLayers, arrayLength);
		break;
	}
	updateGraph();
}

/*!
 * @brief modifies the x-axis scale according to sc
 * @param int sc: new x-axis scale
 *	
 */
void StreamGraph::setXscale(int sc)
{
	scale = sc;
	updateGraph();
}

/*!
* @brief modifies the y-axis scale according to sc
* @param int sc: new y-axis scale
*	
*/
void StreamGraph::setYscale(int sc)
{
	yScale = sc*0.1;
	updateGraph();
}

/*!
 * @brief updates the graph (recalculates the path for each layer) according to specified values
 *	
 */
void StreamGraph::updateGraph()
{
	middleY = 0.5*dimY;
	startX = 0.5*(dimX - scale*arrayLength);
	yMax = 0;
	yMin = dimY;

	QPoint lastPoint;
	QPoint currPoint;
	QPoint c1;
	QPoint c2;

	for(j=0; j<numberOfLayers; j++){

		QPainterPath myPath;
		lastPoint.setX(startX + scale*(0));
		lastPoint.setY(middleY + layers[j].yBottom[0]*yScale);
		if(lastPoint.y() > yMax) yMax = lastPoint.y();
		myPath.moveTo(lastPoint);
		for(int i=0; i<arrayLength-1; i++){
			lastPoint.setX(startX + scale*(i));
			lastPoint.setY(middleY + layers[j].yBottom[i]*yScale);
			currPoint.setX(startX + scale*(i+1));
			currPoint.setY(middleY + layers[j].yBottom[i+1]*yScale);
			c1.setX((lastPoint.x() + currPoint.x())*0.5);
			c1.setY(lastPoint.y());
			c2.setX((lastPoint.x() + currPoint.x())*0.5);
			c2.setY(currPoint.y());
			myPath.cubicTo(c1, c2, currPoint);
			if(lastPoint.y() > yMax) yMax = lastPoint.y();
			if(currPoint.y() > yMax) yMax = currPoint.y();
		}
		lastPoint = currPoint;
		if(lastPoint.y() < yMin) yMin = lastPoint.y();
		currPoint.setX(startX + scale*(arrayLength-1));
		currPoint.setY(middleY + layers[j].yTop[arrayLength-1]*yScale);
		c1.setX(lastPoint.x() );
		c1.setY((lastPoint.y() + currPoint.y())*0.5);
		c2.setX(currPoint.x() );
		c2.setY((lastPoint.y() + currPoint.y())*0.5);
		myPath.cubicTo(c1, c2, currPoint);

		for(int i=arrayLength-1; i>0; i--){
			lastPoint.setX(startX + scale*(i));
			lastPoint.setY(middleY + layers[j].yTop[i]*yScale);
			currPoint.setX(startX + scale*(i-1));
			currPoint.setY(middleY + layers[j].yTop[i-1]*yScale);
			c1.setX((lastPoint.x() + currPoint.x())*0.5);
			c1.setY(lastPoint.y());
			c2.setX((lastPoint.x() + currPoint.x())*0.5);
			c2.setY(currPoint.y());
			myPath.cubicTo(c1, c2, currPoint);
			if(lastPoint.y() < yMin) yMin = lastPoint.y();
			if(currPoint.y() < yMin) yMin = currPoint.y();
		}

		lastPoint = currPoint;
		currPoint.setX(startX + scale*(0));
		currPoint.setY(middleY + layers[j].yBottom[0]*yScale);
		c1.setX(lastPoint.x() );
		c1.setY((lastPoint.y() + currPoint.y())*0.5);
		c2.setX(currPoint.x() );
		c2.setY((lastPoint.y() + currPoint.y())*0.5);
		myPath.cubicTo(c1, c2, currPoint);

		myPath.setFillRule(Qt::OddEvenFill);
		layers[j].path = myPath;
	}
}

/*!
 * @brief returns the graph's layers
 *	
 */
Layer* StreamGraph::getLayers()
{
	return layers;
}

/*!
 * @brief reads a csv file and generates a layer for each artist
 * @param QString filename: the csv filename
 *	
 */
void StreamGraph::loadNewGraph(QString fileName)
{
	vector<string> row;
	string line;
	ifstream in( fileName.toStdString().c_str());
	if (in.fail())  { cout << "File not found" <<endl; }

	int lineCount = 0;
	int weekCount = 0;
	int artistCount = 0;
	int currentweek;
	int lastweek = 0;
	string currentArtist;
	string lastArtist;
	vector<int> weeks;
	vector<string> artists;
	vector<vector <int>> table;
	vector<bool> occurred;

	while(getline(in, line)  && in.good() )
	{
		if(lineCount>0){
			csvline_populate(row, line, ';');
			currentweek = atoi(row.at(0).c_str()); 

			//neue woche gefunden!
			if(currentweek != lastweek){
				//insert 0 for all artists that have not occurred last week
				for(int i=0; i<occurred.size(); i++){
					if(occurred.at(i) == false) table.at(i).push_back(0);
				}

				//no artist has occured yet, because its a new week
				for(int i=0; i<occurred.size(); i++)
					occurred.at(i) = false;
				weekCount++;
				weeks.push_back(currentweek);
				//knstler hinzufgen
				currentArtist = row.at(1);
				//wenn neuer knstler gefunden: array bis hier mit 0 fllen und neuen wert eintragen
				int artistPos = returnIndex(artists, currentArtist);
				if (artistPos == -1){
					artistCount++;
					artists.push_back(currentArtist);
					artistPos = returnIndex(artists, currentArtist);
					table.push_back(vector<int>());
					occurred.push_back(true);
					//fill with 0
					for(int i=0; i<(weekCount-1); i++){
						table.at(artistPos).push_back(0);
					}
					cout << currentArtist << "\t";
				}
				//wenn knstler schon vorhanden:
				table.at(artistPos).push_back(atoi(row.at(2).c_str()));
				try{
					occurred.at(artistPos) = true;
				}catch ( std::out_of_range outOfRange ){
					cout << "\n\nException: 0" << outOfRange.what();	}
			}
			//wenn gleiche woche
			else {
				//restliche knstler der woche hinzufgen
				//wenn vorhandener knstler diese woche nicht dabei: 0 eintragen
				currentArtist = row.at(1);
				//wenn neuer knstler gefunden: array bis hier mit 0 fllen und neuen wert eintragen
				int artistPos = returnIndex(artists, currentArtist);
				if (artistPos == -1){
					artistCount++;
					artists.push_back(currentArtist);
					artistPos = returnIndex(artists, currentArtist);
					table.push_back(vector<int>());
					occurred.push_back(true);
					//fill with 0
					for(int i=0; i<(weekCount-1); i++){
						table.at(artistPos).push_back(0);
					}
					cout << currentArtist << "\t";
				}
				//wenn knstler schon vorhanden:
				table.at(artistPos).push_back(atoi(row.at(2).c_str()));
				try{
					occurred.at(artistPos) = true;
				}catch ( std::out_of_range outOfRange ){
					cout << "\n\nException: 1" << outOfRange.what();}
			}
			lastweek = currentweek;
		}
		cout << endl;
		lineCount++;
	}
	for(int i=0; i<occurred.size(); i++){
		if(occurred.at(i) == false) table.at(i).push_back(0);
	}
	cout << weekCount << "<- weeks || artists ->" << artistCount;
	in.close();

	cout << endl << "tabledata:" << endl;
	for(int i=0; i<table.size(); i++){
		cout << "artist: " << artists[i] << " -> ";
		for(int j=0; j<table.at(i).size(); j++){
			cout << table.at(i).at(j) << "| ";
		}
		cout << endl;
	}

	layers = new Layer[artistCount];
	for (int i=0; i<artistCount; i++){
		int *values = new int[weekCount];
		for (int j=0; j<weekCount; j++){
			values[j] = 10*table[i][j];
		}
		layers[i] = Layer(QString::fromStdString(artists[i]), values, weekCount);
		//layers[i].color = QColor(rand()%255,rand()%255,rand()%255);
	}
	updateColorPalette(colors);

	arrayLength = weekCount;
	numberOfLayers = artistCount;

	setSort(sortOrder);
}

/*!
 * @brief parses a line in the specified csv file
 * @param vector<string> &record: each string from a line is stored as a seperate value in this vektor
 * @param const string& line: the line that is currently being read
 * @param char delimiter: specifies which character seperates strings from each other
 *	
 */
void StreamGraph::csvline_populate(vector<string> &record, const string& line, char delimiter)
{
	int linepos=0;
	int inquotes=false;
	char c;
	int i;
	int linemax=line.length();
	string curstring;
	record.clear();

	while(line[linepos]!=0 && linepos < linemax)
	{

		c = line[linepos];

		if (!inquotes && curstring.length()==0 && c=='"')
		{
			//beginquotechar
			inquotes=true;
		}
		else if (inquotes && c=='"')
		{
			//quotechar
			if ( (linepos+1 <linemax) && (line[linepos+1]=='"') )
			{
				//encountered 2 double quotes in a row (resolves to 1 double quote)
				curstring.push_back(c);
				linepos++;
			}
			else
			{
				//endquotechar
				inquotes=false;
			}
		}
		else if (!inquotes && c==delimiter)
		{
			//end of field
			record.push_back( curstring );
			curstring="";
		}
		else if (!inquotes && (c=='\r' || c=='\n') )
		{
			record.push_back( curstring );
			return;
		}
		else
		{
			curstring.push_back(c);
		}
		linepos++;
	}
	record.push_back( curstring );
	return;
}

/*!
 * @brief returns the index number of a specified artist
 * @param vector<string>vec: contains all artists
 * @param string art: the artist we are looking for
 * @return int: the index of the artist we are looking for, -1 if artist was not found
 */
int StreamGraph::returnIndex(vector<string>vec, string art){
	for (int i = 0; i<vec.size(); i++){
		if(art == vec.at(i))return i;
	}
	//if artist not found
	return -1;
}

/*!
 * @brief sets a layers color according to the specified color array and the coloring criterion (volatility, onset, end, sum)
 * @param float* colors: an array with interpolated colors
 *	
 */
void StreamGraph::updateColorPalette(float* colors)
{
	this->colors = colors;

	float *minmax = getMinMax(criterion);
	cout << "min: " << minmax[0] << " max: " << minmax[1] << endl;
	float range = minmax[1] - minmax[0];
	int step = (int)(4095/range);
	int colIndex;

	for (int i=0; i<numberOfLayers; i++){
		switch (criterion) {
		case 0:
			colIndex = (layers[i].volatility - minmax[0])*step*3;
			break;
		case 1:
			colIndex = (layers[i].onset - minmax[0])*step*3;
			break;
		case 2:
			colIndex = (layers[i].sum - minmax[0])*step*3;
			break;
		case 3:
			colIndex = (layers[i].end - minmax[0])*step*3;
			break;
		}
		
		cout << "col: " << colIndex  << " r:" << colors[colIndex] << " g:" << colors[colIndex+1] << " b:" << colors[colIndex+2] << endl;
		QColor color;
		color.setRedF(colors[colIndex]);
		color.setGreenF(colors[colIndex+1]);
		color.setBlueF(colors[colIndex+2]);
		
		if(twoD){
			float *sum = getMinMaxSum();
			float sumRange = sum[1] - sum[0];
			int sumStep = (int)(sumRange/100);
			float factor = 100+0.5*(layers[i].sum - sum[0])/sumStep;
			cout << "factor = " << factor << endl;
			color = color.lighter(factor);
		}

		layers[i].color = color;
	}
}

/*!
 * @brief calculates minimum and maximum volatility of all layers
 * @return int: returns min and max volatility
 *	
 */
float* StreamGraph::getMinMaxVolatility()
{
	float *minmax =  new float[2];
	minmax[0] = 50000;
	minmax[1] = -1;
	for (int i=0; i<numberOfLayers; i++){
		if(layers[i].volatility < minmax[0]) minmax[0] = layers[i].volatility;
		if(layers[i].volatility > minmax[1]) minmax[1] = layers[i].volatility;
	}

	cout << "min vol: " << minmax[0] << " |max vol: " << minmax[1] << endl;
	return minmax;

}

/*!
* @brief calculates minimum and maximum onset of all layers
* @return int: returns min and max onset
*	
*/
float* StreamGraph::getMinMaxOnset()
{
	float *minmax =  new float[2];
	minmax[0] = 50000;
	minmax[1] = -1;
	for (int i=0; i<numberOfLayers; i++){
		if(layers[i].onset < minmax[0]) minmax[0] = layers[i].onset;
		if(layers[i].onset > minmax[1]) minmax[1] = layers[i].onset;
	}

	cout << "min onset: " << minmax[0] << " |max onset: " << minmax[1] << endl;
	return minmax;
}

/*!
* @brief calculates minimum and maximum end of all layers
* @return int: returns min and max end
*	
*/
float* StreamGraph::getMinMaxEnd()
{
	float *minmax =  new float[2];
	minmax[0] = 50000;
	minmax[1] = -1;
	for (int i=0; i<numberOfLayers; i++){
		if(layers[i].end < minmax[0]) minmax[0] = layers[i].end;
		if(layers[i].end > minmax[1]) minmax[1] = layers[i].end;
	}

	cout << "min end: " << minmax[0] << " |max end: " << minmax[1] << endl;
	return minmax;
}

/*!
* @brief calculates minimum and maximum sum of all layers
* @return int: returns min and max sum
*	
*/
float* StreamGraph::getMinMaxSum()
{
	float *minmax =  new float[2];
	minmax[0] = 50000;
	minmax[1] = -1;
	for (int i=0; i<numberOfLayers; i++){
		if(layers[i].sum < minmax[0]) minmax[0] = layers[i].sum;
		if(layers[i].sum > minmax[1]) minmax[1] = layers[i].sum;
	}
	cout << "min sum: " << minmax[0] << " |max sum: " << minmax[1] << endl;
	return minmax;
}

/*!
 * @brief returns the minimum and maximum of the specified parameter (specified by criterion)
 * @param int crit: criterion (0 = volatility, 1 = onset, 2 = sum, 3 = end)
 * @return float*: min and max value
 *	
 */
float* StreamGraph::getMinMax(int crit)
{
	criterion = crit;
	switch (criterion)
	{
	case 0:
		return getMinMaxVolatility();
	case 1:
		return getMinMaxOnset();
	case 2:
		return getMinMaxSum();
	case 3:
		return getMinMaxEnd();
	}
}