#include <QtGui/QPainter>
#include <QtGui/QPainterPath>
#include <QtGui/QPen>
#include <QtGui/QMessageBox>

#include <iostream> 
#include <fstream>

#include "TFWidget.h"
 
using namespace std; 

/*!
 *	\class TFWidget
	\brief implements the colorpalette for the stream graph.
 */


/*!
 * \brief
 * Initializes TFWidget.
 * 
 * Sets initial polyline at left and right border with black and white as vertex colors.
 * 
 */
TFWidget::TFWidget(QWidget *parent): QWidget(parent)
{
	setPalette(QPalette(QColor(250, 250, 200)));
	setAutoFillBackground(true);

	//initialize polyline
	poly = QPolygon(2);
	areaWidth = 511; //dimension of transfer function
	areaHeight = 31;
	poly.setPoint(0, 0, areaHeight/2);
	poly.setPoint(1, areaWidth, areaHeight/2);
	vertexColors.append(Qt::black);
	vertexColors.append(Qt::white);

	mode = true;  //true if vertex add mode is selected
	drawn = false;
	vertexSelected = false;
	dataSet = false;
	areaSize = 6; //size of click area

	prepareTF();
}

/*!
 * \brief
 * Paints a visualization of interpolated vertex colors.
 * 
 */
void TFWidget::paintEvent(QPaintEvent * /* event */)
{
	QPainter painter(this);
	QPen pen1;			//used to draw borders of set colors
	QPen linePen;	   //linePen is used to visualize the transfer function
	pen1.setWidth(1); 
	linePen.setWidth(1);

	//ofstream out("test2.txt"); 
	//if(!out) { 
	//	cout << "Cannot open file.\n"; 
	//} 

	if (dataSet)
	{
		for (int i=0; i<(areaWidth+1); i++)
		{
			qreal r = data[i*8*3];
			qreal g = data[(i*8*3)+1];
			qreal b = data[(i*8*3)+2];
			//out << i << "R:" << r << ";  G:" << g << ";  B:" << b << "; "/* alpha:" << data[(i*8*4)+3]*/ << endl;

			QColor lineColor;
			lineColor.setRgbF(r, g, b);
			lineColor.getRgbF(&r, &g, &b);
			//out << "2R:" << r << "; 2G:" << g << "; 2B:" << b << endl;
			//QColor lineColor = QColor(data[i*8], data[1+i*8], data[2+i*8], data[3+i*8]);
			linePen.setColor(lineColor);
			painter.setPen(linePen);
			for (int j=0; j<areaHeight; j++)
			{
				painter.drawPoint(i,j);
			}
		}
	}

	//draw borders of the colors that have been selected
	for(int i=0; i<poly.size(); i++)
	{
		pen1.setColor(Qt::black);
		painter.setPen(pen1);

		int vX = poly.at(i).x();
		int vY = poly.at(i).y();
		int length = 2;
		QPolygon vertexArea = QPolygon(4);
		vertexArea.setPoint(0, vX-length, vY-areaHeight);
		vertexArea.setPoint(1, vX-length, vY+areaHeight);
		vertexArea.setPoint(2, vX+length, vY+areaHeight);
		vertexArea.setPoint(3, vX+length, vY-areaHeight);

		painter.drawPolygon(vertexArea);

	}

	//out << "end of texture" << endl << endl;
	//out.close(); 

}

/*!
 * \brief
 * Checks if the click area contains an existing vertex and stores the index of this vertex in poly.
 * 
 * \param event
 * contains information about mousePressEvent.
 * 
 */
void TFWidget::mousePressEvent ( QMouseEvent * event )
{
		QPoint clickPos = event->pos();
		createClickArea(clickPos.x(), clickPos.y());

		//check if a vertex of the polyline lies within the boundaries of the clickedArea polygon
		for (int i=0; i<poly.size(); i++)
		{
			if (clickArea.containsPoint(poly.point(i), Qt::WindingFill))
			{
				vertexSelected = true;
				selectedVertex = poly.point(i);
				selectedIndex = i;
				break;
			}
		}
		update();
}

/*!
 * \brief
 * Draws or deletes vertices according to mode. Edits vertex colors in case of right click.
 * 
 * \param event
 * contains information about mouseReleaseEvent.
 * 
 * If mousePressEvent recognized an existing vertex at click position and the right mouse button was released a QColorDialog is opened and the chosen color replaces the old vertex color.
 * Otherwise a new vertex is inserted at the specified position into the QPolygon poly if vertex-add-mode is active or an existing vertex is removed from poly if delete-mode is active.
 * 
 */
void TFWidget::mouseReleaseEvent ( QMouseEvent * event )
{
	if (event->button() == Qt::RightButton)
	{
		if (vertexSelected)
		{
			vertexColors.replace(selectedIndex, QColorDialog::getColor(Qt::white, this));
			vertexSelected = false;
			update();
		}
		prepareTF();
		return;
	}
	else if(mode)
	{
		tempPoly = poly;
		QPoint newPoint = event->pos();

		//if click position contains an existing vertex, move it to specified position
		if (vertexSelected)
		{
			//assure that vertex can only be moved inside of allowed area
			int y = areaHeight/2;

			if(selectedIndex != 0 && selectedIndex != (poly.size()-1))
			{
				//assure that vertex can only be moved between its predecessor and its successor
				if(newPoint.x() > poly.point((selectedIndex-1)).x() && newPoint.x() < poly.point((selectedIndex+1)).x())
				{
					poly.setPoint(selectedIndex, newPoint.x(), y);
					createClickArea(newPoint.x(), y);
				}
				else if(newPoint.x() < poly.point((selectedIndex-1)).x())
				{
					int x = poly.point((selectedIndex-1)).x()+1;
					poly.setPoint(selectedIndex, x, y);
					createClickArea(x, y);
				}
				else if(newPoint.x() > poly.point((selectedIndex+1)).x())
				{
					int x = poly.point((selectedIndex+1)).x()-1;
					poly.setPoint(selectedIndex, (poly.point((selectedIndex+1)).x()-1), y);
					createClickArea(x, y);
				}
			}
			else if(selectedIndex == 0)
			{
					poly.setPoint(selectedIndex, 0, y);
					createClickArea(0, y);
			}
			else if(selectedIndex == (poly.size()-1))
			{
				poly.setPoint(selectedIndex, areaWidth, y);
				createClickArea(areaWidth, y);
			}

		}
		else
		{
			//call QColorDialog to chose color for new vertex
			QColor color = QColorDialog::getColor(Qt::white, this);

			//if not: create new polygon with +1 vertices
			poly = QPolygon((poly.size()+1));

			//and fill the new polygon sorted by x-value
			for(int i=0; i<tempPoly.size(); i++)
			{
				//if vertex i is bigger than newPoint draw newPoint at i and vertex i at i+1
				if(tempPoly.point(i).x() > newPoint.x() && !drawn)
				{
					poly.setPoint(i, newPoint);
					poly.setPoint((i+1), tempPoly.point(i));
					vertexColors.insert(i, color);
					drawn = true;
				}
				else
				{
					//if newPoint is already drawn, draw remaining vertices at i+1
					if(drawn)
						poly.setPoint((i+1), tempPoly.point(i));
					else
						poly.setPoint(i, tempPoly.point(i));
				}

			}
			createClickArea(newPoint.x(), newPoint.y());
		}
		xcoord = event->pos().x();
		ycoord = event->pos().y();

		update();
		drawn = false;
		vertexSelected = false;
	}
	//remove selected vertex if delete mode is activated
	else if(!mode && selectedIndex != 0 && selectedIndex != (poly.size()-1) && vertexSelected)
	{
		QPoint newPoint = event->pos();
		createClickArea(newPoint.x(), newPoint.y());
		poly.remove(selectedIndex);
		vertexColors.remove(selectedIndex);

		update();
		drawn = false;
		vertexSelected = false;
	}

	prepareTF();
 }

 /*!
 * \brief
 * Sets mode to true so that vertex-add-mode is activated.
 * 
 */
void TFWidget::addMode()
{
	mode = true;
}

/*!
 * \brief
 * Sets mode to false so that vertex-delete-mode is activated.
 * 
 */
void TFWidget::delMode()
{
	mode = false;
}

/*!
 * \brief
 * Resets all parameters in TFWidget to initial values.
 * 
 */
void TFWidget::resetTF()
{
	for(int i = 0; i < 4096; i++){
		data[i] = 0.0f;
	}
	poly = QPolygon(2);
	poly.setPoint(0, 0, areaHeight);
	poly.setPoint(1, areaWidth, 0);
	vertexColors.clear();
	vertexColors.append(Qt::black);
	vertexColors.append(Qt::white);
	createClickArea(0,0);
	prepareTF();
	update();
}

//
/*!
 * \brief
 * Creates polygon around click position with specified size (areaSize).
 * 
 * \param x
 * x-pos around which area is created.
 * 
 * \param y
 * y-pos around which area is created.
 * 
 */
void TFWidget::createClickArea(int x, int y)
{
	clickArea = QPolygon(4);
	clickArea.setPoint(0, x-areaSize, y-areaHeight);
	clickArea.setPoint(1, x-areaSize, y+areaHeight);
	clickArea.setPoint(2, x+areaSize, y+areaHeight);
	clickArea.setPoint(3, x+areaSize, y-areaHeight);
}

/*!
 * \brief
 * Transforms the transfer function into a float array that can be directly used for generating the transfer function texture.
 *  
 * The values between vertices are interpolated and saved to an array that is returned to volGLWidget. 
 * 
 */
void TFWidget::prepareTF()
{
	QPolygonF transfPoly = QPolygonF();

	//ofstream out("test3.txt"); 

	//set correct y-values for vertices
	for(int i=0; i<poly.size(); i++)
	{
		//out << "alpha before: " << poly.point(i).y() << endl;
		float y = 0;

		transfPoly.append(QPointF(poly.point(i).x(), y));
		//out << "alpha: " << y << endl;
	}
	//out.close();

	int dim = 4096;
	int sample = 0;

	//insert 1st RGBA into our texture storage array
	data[0] = vertexColors[sample].redF();
	data[1] = vertexColors[sample].greenF();
	data[2] = vertexColors[sample].blueF();
	int x0 = (transfPoly.at(sample).x())*8;
	int x1 = (transfPoly.at(sample+1).x())*8;

//	ofstream out2("test4.txt"); 
//out2 << "yoyoyo" << endl;
	for(int i = 1; i < dim; i++)
	{
		//if we hit a sample we can write the exact value into the dataset
		if(i == x1)
		{
			sample++;
			data[i*3] = vertexColors[sample].redF();
			data[i*3+1] = vertexColors[sample].greenF();
			data[i*3+2] = vertexColors[sample].blueF();

			/*if(sample == (transfPoly.size()-1))
				break;*/

			x0 = x1;
			x1 = (transfPoly.at(sample+1).x())*8;
		}
		//in every other case we have to interpolate
		else
		{
			data[i*3] = vertexColors[sample].redF() + ((vertexColors[sample+1].redF() - vertexColors[sample].redF())/(x1 - x0)) * (i - x0);
			data[i*3+1] = vertexColors[sample].greenF() + ((vertexColors[sample+1].greenF() - vertexColors[sample].greenF())/(x1 - x0)) * (i - x0);
			data[i*3+2] = vertexColors[sample].blueF() + ((vertexColors[sample+1].blueF() - vertexColors[sample].blueF())/(x1 - x0)) * (i - x0);
		}
		//out2 << "" << i << ": " << data[i*3] << " " << data[i*3+1] << " "	<< data[i*3+2] << endl;
	}
	dataSet = true;

	//out2.close(); 

	emit functionChanged(data);
}

/*!
 * \brief
 * Saves the current transfer function to a file.
 * 
 * \param file
 * points to a QFile object
 *  
 */
void TFWidget::save(QFile* file)
{
	tf.clear();

	for (int i=0; i<poly.size(); i++)
	{
		QVector <int> vertexValues;
		vertexValues.append(poly.point(i).x());
		vertexValues.append(poly.point(i).y());
		vertexValues.append(vertexColors.at(i).red());
		vertexValues.append(vertexColors.at(i).green());
		vertexValues.append(vertexColors.at(i).blue());
		tf.insert(i, vertexValues);
	}

	QDataStream out(file);
	out.setVersion(QDataStream::Qt_4_3);
	out << tf;
}


/*!
 * \brief
 * Loads a transfer function file.
 * 
 * \param file
 * Receives a pointer to a QFile.
 *  
 * A stream on the QFile is opened and loaded onto the QMap tf which stores all the data of the transfer function:
 * vertex positions and their color values.
 * This data is then restored to QPolygon poly and QVector <QColor> vertexColors.
 * Finally the TFWidget gets updated to paint the loaded transfer function. 
 * 
 */
void TFWidget::load(QFile* file)
{
	QDataStream in(file);
	in.setVersion(QDataStream::Qt_4_3);
	tf.clear();   // empty existing tf
	in >> tf;

	if (tf.isEmpty())
	{
		QMessageBox::information(this, tr("No transfer function in file"), tr("The file you are attempting to open contains no transfer function."));
	}
	else
	{
		poly = QPolygon(tf.size());
		vertexColors.clear();

		for (int i=0; i<tf.size(); i++)
		{
			poly.setPoint(i, tf.value(i).at(0), tf.value(i).at(1));
			vertexColors.append(QColor(tf.value(i).at(2), tf.value(i).at(3), tf.value(i).at(4)));
		}

		prepareTF();
		update();
	}

}
